' MD2Plus.DLL
' Library for handling MD2 files via OpenGL
' Version 1.1
'
' Petr Schreiber, 2012
' Use at own risk

' =============================================================================

#COMPILE DLL "MD2PLUS.DLL"
#RESOURCE    "MD2PLUS.res"
#DIM ALL
#COM TLIB OFF

' =============================================================================

' -- OpenGL header by Jos Roca
%USEMACROS = 1
#INCLUDE "opengl32.inc"

' =============================================================================

' -- Type to hold BGRA color data
TYPE BGRA_t
  b AS BYTE
  g AS BYTE
  r AS BYTE
  a AS BYTE
END TYPE

' -- Auxiliary vector types
TYPE vec2_t
  x  AS SINGLE
  y  AS SINGLE
END TYPE

TYPE vec3_t
  x  AS SINGLE
  y  AS SINGLE
  z  AS SINGLE
END TYPE

' -- MD2 related types

TYPE MD2_header_t
  ident             AS LONG    ' magic number: "IDP2"
  version           AS LONG    ' version: must be 8
  skinwidth         AS LONG    ' texture width
  skinheight        AS LONG    ' texture height
  framesize         AS LONG    ' size in bytes of a frame
  num_skins         AS LONG    ' number of skins
  num_vertices      AS LONG    ' number of vertices per frame
  num_st            AS LONG    ' number of texture coordinates
  num_tris          AS LONG    ' number of triangles
  num_glcmds        AS LONG    ' number of opengl commands
  num_frames        AS LONG    ' number of frames
  offset_skins      AS LONG    ' offset skin data
  offset_st         AS LONG    ' offset texture coordinate data
  offset_tris       AS LONG    ' offset triangle data
  offset_frames     AS LONG    ' offset frame data
  offset_glcmds     AS LONG    ' offset OpenGL command data
  offset_end        AS LONG    ' offset end of file
END TYPE

TYPE md2_skin_t
  NAME AS ASCIIZ * 64
END TYPE

TYPE md2_texCoord_t
  s   AS INTEGER
  t   AS INTEGER
END TYPE

TYPE md2_triangle_t
  vertex(1 TO 3) AS WORD
  st(1 TO 3)     AS WORD
END TYPE

TYPE md2_vertex_t
  v(1 TO 3)        AS BYTE
  normalIndex AS BYTE
END TYPE

TYPE md2_frame_t
  SCALE(1 TO 3)      AS SINGLE
  translate(1 TO 3)  AS SINGLE
  NAME          AS ASCIIZ * 16
END TYPE

TYPE tMD2Clone
  Model     AS LONG
  frameFrom AS LONG
  frameTo   AS LONG
  frameNow  AS SINGLE
END TYPE

' =============================================================================

GLOBAL MD2_Normals() AS vec3_t      ' -- Holds precomputed normals
GLOBAL ModelArray() AS iMD2         ' -- Full featured data, precomputed animations
GLOBAL CloneArray() AS tMD2Clone    ' -- Our clones pointing to ModelArrays for rendering

' =============================================================================

MACRO Interpolate(a,b,f) = a + (b-a) * f

' =============================================================================

SUB MD2_Error( message AS STRING )
  MSGBOX message, %MB_ICONINFORMATION, "MD2Plus Error"
END SUB

' =============================================================================

' -- Class holding the heavy data for MD2
CLASS cMD2
  INSTANCE sBuffer AS STRING      ' -- Holds header
  INSTANCE md2Pattern  AS STRING  ' -- Holds info on used file and skin

  INSTANCE texCoord()        AS md2_texCoord_t, _ ' -- Texture coordinates in raw form
           UVs()             AS vec2_t, _         ' -- Texture coordinates in OpenGL form
           triangles()       AS md2_triangle_t, _ ' -- Contains information on trinagles vertices and ST ( UV ) mapping
           frames()          AS md2_frame_t, _    ' -- Info on frame transformations and names
           framesVtx()       AS STRING, _         ' -- Raw frames data buffer
           framesVtxUsable() AS STRING            ' -- Raw data on final OpenGL coordinates evaluated from frames()

'  instance arrayVertex()     as vec3_t, _
'           arrayNormal()     AS vec3_t, _
'           arrayTexUVs()     AS vec2_t, _
'           arrayElements     as long

  INSTANCE textureID AS LONG                      ' -- Handle of OpenGL texure used in this model

  ' -- Object initialization
  CLASS METHOD CREATE()
    sBuffer = STRING$(SIZEOF(MD2_header_t), $SPC)
    REDIM MD2_header(0 TO 0) AS INSTANCE MD2_header_t AT STRPTR(sBuffer)
  END METHOD

  ' -- Object resources deletion
  CLASS METHOD DESTROY()
    glDeleteTextures 1, textureID
  END METHOD

  INTERFACE iMD2
    INHERIT IUNKNOWN

    ' -- Method to load MD2 file and its skin, must be called as first
    METHOD LOAD( fName AS STRING, texName AS STRING ) AS LONG

      REGISTER i AS LONG, j AS LONG
      md2Pattern = UCASE$(fName)+UCASE$(texName)

      sBuffer = FILE_LOAD(fName)
      IF LEN(sBuffer) = 0 THEN MD2_Error(fName+$CRLF+"Invalid file length")    : METHOD = -1 : EXIT METHOD

      ' -- Quickly retrieve header data by overlapping
      REDIM MD2_header(0 TO 0) AS INSTANCE MD2_header_t AT STRPTR(sBuffer)

      IF MD2_header(0).ident    <> 844121161 THEN MD2_Error("Invalid file")    : METHOD = -1 : EXIT METHOD
      IF MD2_header(0).version  <> 8         THEN MD2_Error("Invalid version") : METHOD = -1 : EXIT METHOD

      ' -- Load texture and get its dimensions
      LOCAL w, h AS LONG

      glGenTextures 1, textureID
      TextureLoadBMP(texName, textureID, w, h)

      ' -- Memory allocation
      REDIM texCoord (1 TO MAX(MD2_header(0).num_st    , 1))
      REDIM UVs      (1 TO MAX(MD2_header(0).num_st    , 1))
      REDIM triangles(1 TO MAX(MD2_header(0).num_tris  , 1))
      REDIM frames   (1 TO MAX(MD2_header(0).num_frames, 1))
      REDIM framesVtx(1 TO MAX(MD2_header(0).num_frames, 1))

      REDIM framesVtxUsable(1 TO MAX(MD2_header(0).num_frames, 1))

      ' -- Esoteric handling of offsets ;)
      DECR MD2_header(0).offset_skins
      INCR MD2_header(0).offset_st
      INCR MD2_header(0).offset_tris
      DECR MD2_header(0).offset_frames

      ' -- Retrieving ST (UV) data and triangle data
      IF MD2_header(0).num_st     THEN POKE$ VARPTR(texCoord(1).s)         , MID$(sBuffer, MD2_header(0).offset_st,     MD2_header(0).num_st * SIZEOF(md2_texCoord_t) )
      IF MD2_header(0).num_tris   THEN POKE$ VARPTR(triangles(1).vertex(1)), MID$(sBuffer, MD2_header(0).offset_tris,   MD2_header(0).num_tris * SIZEOF(md2_triangle_t) )

      ' -- Indices are zero based, but I hate it so I will INCR it by one
      FOR i = 1 TO MD2_header(0).num_tris
        triangles(i).vertex(1) += 1   ' -- Coordinates for each triangle
        triangles(i).vertex(2) += 1
        triangles(i).vertex(3) += 1

        triangles(i).st(1) += 1       ' -- UVs for each triangle
        triangles(i).st(2) += 1
        triangles(i).st(3) += 1
      NEXT

      FOR i = 1 TO MD2_header(0).num_st
        UVs(i).x =  texCoord(i).s / w  ' -- "Well done" coordinates according to real texture size
        UVs(i).y = -texCoord(i).t / h
      NEXT

      ' -- Retrieving frame data
      GLOBAL FrameChunkLenA AS LONG
      FrameChunkLenA = SIZEOF(md2_frame_t)

      GLOBAL FrameChunkLenB AS LONG
      FrameChunkLenB = SIZEOF(md2_vertex_t) * MD2_header(0).num_vertices

      GLOBAL FrameChunkPos AS LONG
      FrameChunkPos = MD2_header(0).offset_frames+2

      GLOBAL FrameChunkA, FrameChunkB AS STRING

      FOR i = 1 TO MD2_header(0).num_frames
        FrameChunkA = MID$(sBuffer, FrameChunkPos, FrameChunkLenA)
        FrameChunkB = MID$(sBuffer, FrameChunkPos+FrameChunkLenA, FrameChunkLenB)

        POKE$ VARPTR(frames(i).SCALE(1)), FrameChunkA

        framesVtx(i) = FrameChunkB
        FrameChunkPos += (FrameChunkLenA+FrameChunkLenB)

      NEXT

      ' -- Precomputing frames saves lot of processing time
      LOCAL n AS LONG
      FOR n = 1 TO MD2_header(0).num_frames
        framesVtxUsable(n) = STRING$(SIZEOF(vec3_t)*MD2_header(0).num_vertices, $SPC)
        REDIM vertex(1 TO MD2_header(0).num_vertices ) AS vec3_t AT STRPTR(framesVtxUsable(n))
        REDIM FrameData(1 TO MD2_header(0).num_vertices) AS md2_vertex_t AT STRPTR(framesVtx(n))

        FOR j = 1 TO MD2_header(0).num_vertices
          vertex(j).x = (frames(n).SCALE(1) * FrameData(j).v(1)) + frames(n).translate(1)
          vertex(j).y = (frames(n).SCALE(2) * FrameData(j).v(2)) + frames(n).translate(2)
          vertex(j).z = (frames(n).SCALE(3) * FrameData(j).v(3)) + frames(n).translate(3)
        NEXT
      NEXT

'      ' -- Arrays for VA rendering
'      dim arrayIndex    as long
'      dim indexST       as long
'      arrayElements = MD2_header(0).num_tris * 3
'      arrayIndex    = 0
'
'      redim arrayVertex(1 TO MAX(arrayElements, 1))
'      redim arrayNormal(1 TO MAX(arrayElements, 1))
'      redim arrayTexUVs(1 TO MAX(arrayElements, 1))
'
'      ' -- Cache UVs
'      FOR i = 1 TO MD2_header(0).num_tris
'        ' -- draw each vertex
'        FOR j = 1 TO 3
'          ' -- calculate vertex data
'          INCR arrayIndex
'
'          indexST  = triangles(i).st(j)
'
'          TYPE SET arrayTexUVs(arrayIndex) = UVs(indexST)
'        NEXT
'      NEXT

    END METHOD

    ' -- Draws interpolation between two frames or just that keyframe
    METHOD DRAW(n AS SINGLE, suggestedNext AS LONG)

      REGISTER i AS LONG, j AS LONG

      DIM v(1 TO 3) AS SINGLE
      DIM interpol_a(1 TO 3) AS SINGLE
      DIM interpol_b(1 TO 3) AS SINGLE
      DIM interpol_n(1 TO 3) AS SINGLE

      ' -- Floating and integer part of frame
      LOCAL fp AS SINGLE
      LOCAL ip AS SINGLE

      fp = FRAC(n)
      ip = FIX(n)

      DIM FrameData(1 TO MD2_header(0).num_vertices) AS md2_vertex_t AT STRPTR(framesVtx(ip))
      REDIM vertex(1 TO MD2_header(0).num_vertices ) AS vec3_t AT STRPTR(framesVtxUsable(ip))

      LOCAL indexPos, indexST, indexN, indexN2 AS LONG
      LOCAL disableTex AS LONG
      LOCAL nextFrame AS LONG

      nextFrame = IIF(suggestedNext > 0, suggestedNext, ip+1)

      ' -- Next frame data
      IF nextFrame <= MD2_header(0).num_frames THEN
        REDIM FrameDataNext(1 TO MD2_header(0).num_vertices) AS md2_vertex_t AT STRPTR(framesVtx(nextFrame))
        REDIM vertexNext(1 TO MD2_header(0).num_vertices ) AS vec3_t AT STRPTR(framesVtxUsable(nextFrame))
      END IF

      glPushMatrix
        IF glIsEnabled(%GL_TEXTURE_2D) = 0 THEN
          glEnable %GL_TEXTURE_2D
          disableTex = 1
        END IF
        glBindTexture %GL_TEXTURE_2D, textureID

        ' -- Rotates model in such a way Y points up
        glRotatef -90,1,0,0
        glRotatef -90,0,0,1

'        ' -- Rendering via Vertex Arrays
'        ' -- Enable array mapping
'        glEnableClientState( %GL_NORMAL_ARRAY )
'        glNormalPointer  (%GL_FLOAT, SIZEOF(vec3_t), arrayNormal(1))
'
'        glEnableClientState( %GL_TEXTURE_COORD_ARRAY )
'        glTexCoordPointer( 2, %GL_FLOAT, SIZEOF(vec2_t), arrayTexUVs(1))
'
'        glEnableClientState( %GL_VERTEX_ARRAY )
'        glVertexPointer  ( 3, %GL_FLOAT, SIZEOF(vec3_t), arrayVertex(1))
'
'        ' -- Our index for VA arrays
'        dim arrayIndex as long
'        arrayIndex = 0
'
'        ' -- No interpolation involved?
'        IF fp = 0 OR nextFrame > MD2_header(0).num_frames THEN
'
'
'          ' -- draw each triangle
'          FOR i = 1 TO MD2_header(0).num_tris
'            ' -- draw each vertex
'            FOR j = 1 TO 3
'              ' -- calculate vertex data
'              incr arrayIndex
'
'              indexPos = triangles(i).vertex(j)
'              indexN   = FrameData(indexPos).normalIndex+1
'
'              type set arrayNormal(arrayIndex) = MD2_Normals(indexN)
'              TYPE SET arrayVertex(arrayIndex) = vertex(indexPos)
'            NEXT
'          NEXT
'
'        ELSE
'        ' -- Rendering interpolation
'        ' -- draw each triangle
'          FOR i = 1 TO MD2_header(0).num_tris
'            ' -- draw each vertex
'            FOR j = 1 TO 3
'              INCR arrayIndex
'
'              indexPos = triangles(i).vertex(j)
'              indexN   = FrameData(indexPos).normalIndex+1
'              indexN2  = FrameDataNext(indexPos).normalIndex +1
'
'              ' -- Array assign is the fastest way to assign to ... arrays! :)
'              ARRAY ASSIGN v() = Interpolate(vertex(indexPos).x, vertexNext(indexPos).x, fp), _
'                                 Interpolate(vertex(indexPos).y, vertexNext(indexPos).y, fp), _
'                                 Interpolate(vertex(indexPos).z, vertexNext(indexPos).z, fp)
'
'              ' --
'
'              ARRAY ASSIGN interpol_n() = Interpolate(MD2_Normals(indexN).x, MD2_Normals(indexN2).x, fp), _
'                                          Interpolate(MD2_Normals(indexN).y, MD2_Normals(indexN2).y, fp), _
'                                          Interpolate(MD2_Normals(indexN).z, MD2_Normals(indexN2).z, fp)
'
'              ' --
'
'              type set arrayNormal(arrayIndex) = peek$(varptr(interpol_n(1)), sizeof(vec3_t))
'              TYPE SET arrayVertex(arrayIndex) = PEEK$(VARPTR(v(1)), SIZEOF(vec3_t))
'            NEXT
'          NEXT
'        END IF
'        ' -- Render it!
'        glDrawArrays(%GL_TRIANGLES, 0, arrayElements)
'
'        ' -- Disable array mapping
'        glDisableClientState( %GL_VERTEX_ARRAY )
'        glDisableClientState( %GL_TEXTURE_COORD_ARRAY )
'        glDisableClientState( %GL_NORMAL_ARRAY )

        ' -- Rendering via immediate mode - this part cries for VA or VBO optimization
        glBegin %GL_TRIANGLES

        ' -- No interpolation involved?
        IF fp = 0 OR nextFrame > MD2_header(0).num_frames THEN
          ' -- draw each triangle
          FOR i = 1 TO MD2_header(0).num_tris
            ' -- draw each vertex
            FOR j = 1 TO 3
              ' -- calculate vertex data
              indexPos = triangles(i).vertex(j)
              indexST  = triangles(i).st(j)
              indexN   = FrameData(indexPos).normalIndex+1

              glNormal3fv MD2_Normals(indexN).x
              glTexCoord2fv UVs(indexST).x
              glVertex3fv vertex(indexPos).x
            NEXT
          NEXT

        ELSE
        ' -- Rendering interpolation
        ' -- draw each triangle
          FOR i = 1 TO MD2_header(0).num_tris
            ' -- draw each vertex
            FOR j = 1 TO 3

              indexPos = triangles(i).vertex(j)
              indexST  = triangles(i).st(j)
              indexN   = FrameData(indexPos).normalIndex+1
              indexN2  = FrameDataNext(indexPos).normalIndex +1

              ' -- Array assign is the fastest way to assign to ... arrays! :)
              ARRAY ASSIGN v() = Interpolate(vertex(indexPos).x, vertexNext(indexPos).x, fp), _
                                 Interpolate(vertex(indexPos).y, vertexNext(indexPos).y, fp), _
                                 Interpolate(vertex(indexPos).z, vertexNext(indexPos).z, fp)

              ' --

              ARRAY ASSIGN interpol_n() = Interpolate(MD2_Normals(indexN).x, MD2_Normals(indexN2).x, fp), _
                                          Interpolate(MD2_Normals(indexN).y, MD2_Normals(indexN2).y, fp), _
                                          Interpolate(MD2_Normals(indexN).z, MD2_Normals(indexN2).z, fp)

              ' --

              glNormal3fv interpol_n(1)
              glTexCoord2fv UVs(indexST).x
              glVertex3fv v(1)
            NEXT
          NEXT
        END IF
        glEnd

        IF disableTex = 1 THEN glDisable %GL_TEXTURE_2D
      glPopMatrix

    END METHOD

    ' -- When you need to get frame name
    PROPERTY GET FrameName( BYVAL n AS LONG ) AS STRING
      PROPERTY = frames(n).NAME
    END PROPERTY

    ' -- When you need to get frame count
    PROPERTY GET GetFrameCount() AS LONG
      PROPERTY = MD2_header(0).num_frames
    END PROPERTY

    ' -- Returns UCASE$ of model file and skin ( ucasing done during loading )
    METHOD Pattern() AS STRING
      METHOD = md2Pattern
    END METHOD

  END INTERFACE
END CLASS

' =============================================================================
' -- What follows are functions to manage just animation data

SUB LockAnimRange( mItem AS tMD2Clone, aFrom AS LONG, aTo AS LONG )
  mItem.frameNow  = aFrom
  mItem.frameFrom = aFrom
  mItem.frameTo   = aTo
END SUB

SUB UnLockAnimRange(mItem AS tMD2Clone)
  mItem.frameNow  = 0
  mItem.frameFrom = 0
  mItem.frameTo   = 0
END SUB

FUNCTION AdvanceAnim(mItem AS tMD2Clone, factor AS SINGLE) AS LONG
 LOCAL nextFrame AS LONG
  mItem.frameNow += factor

  ' -- Forward
  IF factor > 0 THEN

    IF FIX(mItem.frameNow) = mItem.frameTo THEN
      nextFrame = mItem.frameFrom
    ELSE
      nextFrame = FIX(mItem.frameNow)+1
    END IF

    IF mItem.frameNow > mItem.frameTo+1 THEN mItem.frameNow = mItem.frameFrom + FRAC(mItem.frameNow)

  ELSE
  ' -- Backward
    IF mItem.frameNow < mItem.frameFrom THEN mItem.frameNow = mItem.frameTo + FRAC(mItem.frameNow)

    IF FIX(mItem.frameNow) = mItem.frameTo THEN
      nextFrame = mItem.frameFrom
    ELSE
      nextFrame = FIX(mItem.frameNow)+1
    END IF

  END IF

  FUNCTION = nextFrame
END FUNCTION

SUB DrawAnim( mItem AS tMD2Clone, factor AS SINGLE )
  LOCAL nextFrame AS LONG

  nextFrame = AdvanceAnim(mItem, factor)

  ' -- Render result
  ModelArray(mItem.Model).Draw(mItem.frameNow, nextFrame)

END SUB

' =============================================================================
' =============================================================================
' =============================================================================
' -- Exported functions

' -- Loads MD2 file
FUNCTION MD2_Load ALIAS "MD2_Load" ( BYVAL index AS LONG, BYVAL fName AS STRING, BYVAL texName AS STRING ) EXPORT AS LONG
  REGISTER i AS LONG
  LOCAL found, failure AS LONG

  fName   = UCASE$(fName)
  texName = UCASE$(texName)

  ' -- Did we uploaded master already?
  FOR i = 1 TO UBOUND(ModelArray)
    IF ISOBJECT(ModelArray(i)) THEN
      IF ModelArray(i).Pattern = fName+texName THEN
        found = i
        EXIT FOR
      END IF
    END IF
  NEXT

  ' -- If not
  IF found = 0 THEN

    FOR i = 1 TO UBOUND(ModelArray)    ' -- Check if some slot is free and refill it
      IF ISNOTHING(ModelArray(i)) THEN
        found = i
        ModelArray(found) = CLASS "cMD2"
        failure = ModelArray(found).LOAD(fName, texName)
        EXIT FOR
      END IF
    NEXT

    ' -- No slot free, we need to expand
    IF found = 0 THEN
      found = UBOUND(ModelArray)+1
      REDIM PRESERVE ModelArray(1 TO found )
      ModelArray(found) = CLASS "cMD2"
      failure = ModelArray(found).LOAD(fName, texName)
    END IF

  END IF

  ' -- Found is <> 0 for sure now
  IF failure THEN FUNCTION = -1 : EXIT FUNCTION

  IF index > UBOUND(CloneArray) THEN REDIM PRESERVE CloneArray(1 TO index )


  ' -- Which model we will render
  CloneArray(index).Model = found

  FUNCTION = 0

END FUNCTION

' -- Draws MD2 file
FUNCTION MD2_Draw ALIAS "MD2_Draw" ( BYVAL index AS LONG, BYVAL n AS SINGLE) EXPORT AS LONG
  ModelArray(CloneArray(index).Model).Draw(n, -1)
END FUNCTION

' --
FUNCTION MD2_LockAnimRangeByName ALIAS "MD2_LockAnimRangeByName" ( BYVAL index AS LONG, BYVAL s AS STRING ) EXPORT AS LONG
  REGISTER i AS LONG
  LOCAL a, b AS LONG

  s = UCASE$(s)

  LOCAL MaxFrames AS LONG

  MaxFrames = ModelArray(CloneArray(index).Model).GetFrameCount

  FOR i = 1 TO MaxFrames

    ' -- The following line removes any numbers from frame names, so we can compare just animation pattern
    IF REMOVE$(UCASE$(ModelArray(CloneArray(index).Model).FrameName(i)), ANY "1234567890") = s THEN
      IF a = 0 THEN
        a = i
        b = i
      ELSE
        b = i
      END IF
    END IF

  NEXT

  ' -- a = 0 means no such a animation
  IF a = 0 THEN
    FUNCTION = -1
  ELSE
    LockAnimRange(CloneArray(index), a, b)
  END IF

END FUNCTION

' --
FUNCTION MD2_LockAnimRange ALIAS "MD2_LockAnimRange" ( BYVAL index AS LONG, BYVAL a AS LONG, BYVAL b AS LONG ) EXPORT AS LONG
  LockAnimRange(CloneArray(index), a, b)
END FUNCTION

' --
FUNCTION MD2_UnLockAnimRange ALIAS "MD2_UnLockAnimRange" (BYVAL index AS LONG) EXPORT AS LONG
  UnLockAnimRange(CloneArray(index))
END FUNCTION

' --
FUNCTION MD2_AdvanceAnim ALIAS "MD2_AdvanceAnim" (BYVAL index AS LONG, BYVAL n AS SINGLE) EXPORT AS LONG
  AdvanceAnim(CloneArray(index), n)
END FUNCTION

' --
FUNCTION MD2_DrawAnim ALIAS "MD2_DrawAnim" (BYVAL index AS LONG, BYVAL n AS SINGLE) EXPORT AS LONG
  DrawAnim(CloneArray(index), n)
END FUNCTION

' --
FUNCTION MD2_Delete ALIAS "MD2_Delete" (BYVAL index AS LONG) EXPORT AS LONG
  REGISTER i AS LONG
  LOCAL pParent AS LONG
  LOCAL found AS LONG

  pParent = CloneArray(index).Model
  RESET CloneArray(index)

  ' -- We check if PBMB ( Parent Big Model Buffer ) is used by other clones
  FOR i = 1 TO UBOUND(CloneArray)
    IF CloneArray(i).Model = pParent THEN found = 1
  NEXT

  ' -- If not, lets remove that fat data
  IF found = 0 THEN
    ModelArray(pParent) = NOTHING
  END IF

END FUNCTION

' -- PB garbage collection should handle this
FUNCTION MD2_DeleteAll ALIAS "MD2_DeleteAll" () EXPORT AS LONG
  REGISTER i AS LONG
  REDIM CloneArray(1)

  ' -- Releasing objects this way might not be necessary, REDIM should do it, but for better feeling...
  FOR i = 1 TO UBOUND(ModelArray)
    ModelArray(i) = NOTHING
  NEXT

  REDIM ModelArray(1)
END FUNCTION

' --
FUNCTION MD2_GetFrameCount ALIAS "MD2_GetFrameCount" (BYVAL index AS LONG) EXPORT AS LONG
  FUNCTION = ModelArray(CloneArray(index).Model).GetFrameCount
END FUNCTION

' --
FUNCTION MD2_GetFrameList ALIAS "MD2_GetFrameList" (BYVAL index AS LONG) EXPORT AS STRING
  REGISTER i AS LONG
  DIM s AS STRING

  FOR i = 1 TO ModelArray(CloneArray(index).Model).GetFrameCount
    s += ModelArray(CloneArray(index).Model).FrameName(i)+","
  NEXT
  s = RTRIM$(s, ",")
  FUNCTION = s
END FUNCTION

'--
FUNCTION MD2_GetCurrentFrame ALIAS "MD2_GetCurrentFrame" (BYVAL index AS LONG) EXPORT AS SINGLE
  FUNCTION = CloneArray(index).frameNow
END FUNCTION

' =============================================================================
' -- Main DLL entry point called by Windows...

FUNCTION LIBMAIN (BYVAL hInstance   AS LONG, _
                  BYVAL fwdReason   AS LONG, _
                  BYVAL lpvReserved AS LONG) AS LONG

    SELECT CASE fwdReason

    CASE %DLL_PROCESS_ATTACH
        'Indicates that the DLL is being loaded by another process (a DLL
        'or EXE is loading the DLL).  DLLs can use this opportunity to
        'initialize any instance or global data, such as arrays.

        DIM ModelArray(1 TO 1) AS GLOBAL iMD2
        DIM CloneArray(1 TO 1) AS GLOBAL tMD2Clone
        DIM MD2_Normals(1 TO 162) AS GLOBAL vec3_t

        ' -- Precomputed normals
        MD2_Normals(1).x =  -0.525731
        MD2_Normals(1).y =   0.000000
        MD2_Normals(1).z =   0.850651
        MD2_Normals(2).x =   -0.442863
        MD2_Normals(2).y =   0.238856
        MD2_Normals(2).z =   0.864188
        MD2_Normals(3).x =   -0.295242
        MD2_Normals(3).y =   0.000000
        MD2_Normals(3).z =   0.955423
        MD2_Normals(4).x =   -0.309017
        MD2_Normals(4).y =   0.500000
        MD2_Normals(4).z =   0.809017
        MD2_Normals(5).x =   -0.162460
        MD2_Normals(5).y =   0.262866
        MD2_Normals(5).z =   0.951056
        MD2_Normals(6).x =    0.000000
        MD2_Normals(6).y =   0.000000
        MD2_Normals(6).z =   1.000000
        MD2_Normals(7).x =    0.000000
        MD2_Normals(7).y =   0.850651
        MD2_Normals(7).z =   0.525731
        MD2_Normals(8).x =   -0.147621
        MD2_Normals(8).y =   0.716567
        MD2_Normals(8).z =   0.681718
        MD2_Normals(9).x =    0.147621
        MD2_Normals(9).y =   0.716567
        MD2_Normals(9).z =   0.681718
        MD2_Normals(10).x =    0.000000
        MD2_Normals(10).y =   0.525731
        MD2_Normals(10).z =   0.850651
        MD2_Normals(11).x =    0.309017
        MD2_Normals(11).y =   0.500000
        MD2_Normals(11).z =   0.809017
        MD2_Normals(12).x =    0.525731
        MD2_Normals(12).y =   0.000000
        MD2_Normals(12).z =   0.850651
        MD2_Normals(13).x =    0.295242
        MD2_Normals(13).y =   0.000000
        MD2_Normals(13).z =   0.955423
        MD2_Normals(14).x =    0.442863
        MD2_Normals(14).y =   0.238856
        MD2_Normals(14).z =   0.864188
        MD2_Normals(15).x =    0.162460
        MD2_Normals(15).y =   0.262866
        MD2_Normals(15).z =   0.951056
        MD2_Normals(16).x =   -0.681718
        MD2_Normals(16).y =   0.147621
        MD2_Normals(16).z =   0.716567
        MD2_Normals(17).x =   -0.809017
        MD2_Normals(17).y =   0.309017
        MD2_Normals(17).z =   0.500000
        MD2_Normals(18).x =   -0.587785
        MD2_Normals(18).y =   0.425325
        MD2_Normals(18).z =   0.688191
        MD2_Normals(19).x =   -0.850651
        MD2_Normals(19).y =   0.525731
        MD2_Normals(19).z =   0.000000
        MD2_Normals(20).x =   -0.864188
        MD2_Normals(20).y =   0.442863
        MD2_Normals(20).z =   0.238856
        MD2_Normals(21).x =   -0.716567
        MD2_Normals(21).y =   0.681718
        MD2_Normals(21).z =   0.147621
        MD2_Normals(22).x =   -0.688191
        MD2_Normals(22).y =   0.587785
        MD2_Normals(22).z =   0.425325
        MD2_Normals(23).x =   -0.500000
        MD2_Normals(23).y =   0.809017
        MD2_Normals(23).z =   0.309017
        MD2_Normals(24).x =   -0.238856
        MD2_Normals(24).y =   0.864188
        MD2_Normals(24).z =   0.442863
        MD2_Normals(25).x =   -0.425325
        MD2_Normals(25).y =   0.688191
        MD2_Normals(25).z =   0.587785
        MD2_Normals(26).x =   -0.716567
        MD2_Normals(26).y =   0.681718
        MD2_Normals(26).z =  -0.147621
        MD2_Normals(27).x =   -0.500000
        MD2_Normals(27).y =   0.809017
        MD2_Normals(27).z =  -0.309017
        MD2_Normals(28).x =   -0.525731
        MD2_Normals(28).y =   0.850651
        MD2_Normals(28).z =   0.000000
        MD2_Normals(29).x =    0.000000
        MD2_Normals(29).y =   0.850651
        MD2_Normals(29).z =  -0.525731
        MD2_Normals(30).x =   -0.238856
        MD2_Normals(30).y =   0.864188
        MD2_Normals(30).z =  -0.442863
        MD2_Normals(31).x =    0.000000
        MD2_Normals(31).y =   0.955423
        MD2_Normals(31).z =  -0.295242
        MD2_Normals(32).x =   -0.262866
        MD2_Normals(32).y =   0.951056
        MD2_Normals(32).z =  -0.162460
        MD2_Normals(33).x =    0.000000
        MD2_Normals(33).y =   1.000000
        MD2_Normals(33).z =   0.000000
        MD2_Normals(34).x =    0.000000
        MD2_Normals(34).y =   0.955423
        MD2_Normals(34).z =   0.295242
        MD2_Normals(35).x =   -0.262866
        MD2_Normals(35).y =   0.951056
        MD2_Normals(35).z =   0.162460
        MD2_Normals(36).x =    0.238856
        MD2_Normals(36).y =   0.864188
        MD2_Normals(36).z =   0.442863
        MD2_Normals(37).x =    0.262866
        MD2_Normals(37).y =   0.951056
        MD2_Normals(37).z =   0.162460
        MD2_Normals(38).x =    0.500000
        MD2_Normals(38).y =   0.809017
        MD2_Normals(38).z =   0.309017
        MD2_Normals(39).x =    0.238856
        MD2_Normals(39).y =   0.864188
        MD2_Normals(39).z =  -0.442863
        MD2_Normals(40).x =    0.262866
        MD2_Normals(40).y =   0.951056
        MD2_Normals(40).z =  -0.162460
        MD2_Normals(41).x =    0.500000
        MD2_Normals(41).y =   0.809017
        MD2_Normals(41).z =  -0.309017
        MD2_Normals(42).x =    0.850651
        MD2_Normals(42).y =   0.525731
        MD2_Normals(42).z =   0.000000
        MD2_Normals(43).x =    0.716567
        MD2_Normals(43).y =   0.681718
        MD2_Normals(43).z =   0.147621
        MD2_Normals(44).x =    0.716567
        MD2_Normals(44).y =   0.681718
        MD2_Normals(44).z =  -0.147621
        MD2_Normals(45).x =    0.525731
        MD2_Normals(45).y =   0.850651
        MD2_Normals(45).z =   0.000000
        MD2_Normals(46).x =    0.425325
        MD2_Normals(46).y =   0.688191
        MD2_Normals(46).z =   0.587785
        MD2_Normals(47).x =    0.864188
        MD2_Normals(47).y =   0.442863
        MD2_Normals(47).z =   0.238856
        MD2_Normals(48).x =    0.688191
        MD2_Normals(48).y =   0.587785
        MD2_Normals(48).z =   0.425325
        MD2_Normals(49).x =    0.809017
        MD2_Normals(49).y =   0.309017
        MD2_Normals(49).z =   0.500000
        MD2_Normals(50).x =    0.681718
        MD2_Normals(50).y =   0.147621
        MD2_Normals(50).z =   0.716567
        MD2_Normals(51).x =    0.587785
        MD2_Normals(51).y =   0.425325
        MD2_Normals(51).z =   0.688191
        MD2_Normals(52).x =    0.955423
        MD2_Normals(52).y =   0.295242
        MD2_Normals(52).z =   0.000000
        MD2_Normals(53).x =    1.000000
        MD2_Normals(53).y =   0.000000
        MD2_Normals(53).z =   0.000000
        MD2_Normals(54).x =    0.951056
        MD2_Normals(54).y =   0.162460
        MD2_Normals(54).z =   0.262866
        MD2_Normals(55).x =    0.850651
        MD2_Normals(55).y =  -0.525731
        MD2_Normals(55).z =   0.000000
        MD2_Normals(56).x =    0.955423
        MD2_Normals(56).y =  -0.295242
        MD2_Normals(56).z =   0.000000
        MD2_Normals(57).x =    0.864188
        MD2_Normals(57).y =  -0.442863
        MD2_Normals(57).z =   0.238856
        MD2_Normals(58).x =    0.951056
        MD2_Normals(58).y =  -0.162460
        MD2_Normals(58).z =   0.262866
        MD2_Normals(59).x =    0.809017
        MD2_Normals(59).y =  -0.309017
        MD2_Normals(59).z =   0.500000
        MD2_Normals(60).x =    0.681718
        MD2_Normals(60).y =  -0.147621
        MD2_Normals(60).z =   0.716567
        MD2_Normals(61).x =    0.850651
        MD2_Normals(61).y =   0.000000
        MD2_Normals(61).z =   0.525731
        MD2_Normals(62).x =    0.864188
        MD2_Normals(62).y =   0.442863
        MD2_Normals(62).z =  -0.238856
        MD2_Normals(63).x =    0.809017
        MD2_Normals(63).y =   0.309017
        MD2_Normals(63).z =  -0.500000
        MD2_Normals(64).x =    0.951056
        MD2_Normals(64).y =   0.162460
        MD2_Normals(64).z =  -0.262866
        MD2_Normals(65).x =    0.525731
        MD2_Normals(65).y =   0.000000
        MD2_Normals(65).z =  -0.850651
        MD2_Normals(66).x =    0.681718
        MD2_Normals(66).y =   0.147621
        MD2_Normals(66).z =  -0.716567
        MD2_Normals(67).x =    0.681718
        MD2_Normals(67).y =  -0.147621
        MD2_Normals(67).z =  -0.716567
        MD2_Normals(68).x =    0.850651
        MD2_Normals(68).y =   0.000000
        MD2_Normals(68).z =  -0.525731
        MD2_Normals(69).x =    0.809017
        MD2_Normals(69).y =  -0.309017
        MD2_Normals(69).z =  -0.500000
        MD2_Normals(70).x =    0.864188
        MD2_Normals(70).y =  -0.442863
        MD2_Normals(70).z =  -0.238856
        MD2_Normals(71).x =    0.951056
        MD2_Normals(71).y =  -0.162460
        MD2_Normals(71).z =  -0.262866
        MD2_Normals(72).x =    0.147621
        MD2_Normals(72).y =   0.716567
        MD2_Normals(72).z =  -0.681718
        MD2_Normals(73).x =    0.309017
        MD2_Normals(73).y =   0.500000
        MD2_Normals(73).z =  -0.809017
        MD2_Normals(74).x =    0.425325
        MD2_Normals(74).y =   0.688191
        MD2_Normals(74).z =  -0.587785
        MD2_Normals(75).x =    0.442863
        MD2_Normals(75).y =   0.238856
        MD2_Normals(75).z =  -0.864188
        MD2_Normals(76).x =    0.587785
        MD2_Normals(76).y =   0.425325
        MD2_Normals(76).z =  -0.688191
        MD2_Normals(77).x =    0.688191
        MD2_Normals(77).y =   0.587785
        MD2_Normals(77).z =  -0.425325
        MD2_Normals(78).x =   -0.147621
        MD2_Normals(78).y =   0.716567
        MD2_Normals(78).z =  -0.681718
        MD2_Normals(79).x =   -0.309017
        MD2_Normals(79).y =   0.500000
        MD2_Normals(79).z =  -0.809017
        MD2_Normals(80).x =    0.000000
        MD2_Normals(80).y =   0.525731
        MD2_Normals(80).z =  -0.850651
        MD2_Normals(81).x =   -0.525731
        MD2_Normals(81).y =   0.000000
        MD2_Normals(81).z =  -0.850651
        MD2_Normals(82).x =   -0.442863
        MD2_Normals(82).y =   0.238856
        MD2_Normals(82).z =  -0.864188
        MD2_Normals(83).x =   -0.295242
        MD2_Normals(83).y =   0.000000
        MD2_Normals(83).z =  -0.955423
        MD2_Normals(84).x =   -0.162460
        MD2_Normals(84).y =   0.262866
        MD2_Normals(84).z =  -0.951056
        MD2_Normals(85).x =    0.000000
        MD2_Normals(85).y =   0.000000
        MD2_Normals(85).z =  -1.000000
        MD2_Normals(86).x =    0.295242
        MD2_Normals(86).y =   0.000000
        MD2_Normals(86).z =  -0.955423
        MD2_Normals(87).x =    0.162460
        MD2_Normals(87).y =   0.262866
        MD2_Normals(87).z =  -0.951056
        MD2_Normals(88).x =   -0.442863
        MD2_Normals(88).y =  -0.238856
        MD2_Normals(88).z =  -0.864188
        MD2_Normals(89).x =   -0.309017
        MD2_Normals(89).y =  -0.500000
        MD2_Normals(89).z =  -0.809017
        MD2_Normals(90).x =   -0.162460
        MD2_Normals(90).y =  -0.262866
        MD2_Normals(90).z =  -0.951056
        MD2_Normals(91).x =    0.000000
        MD2_Normals(91).y =  -0.850651
        MD2_Normals(91).z =  -0.525731
        MD2_Normals(92).x =   -0.147621
        MD2_Normals(92).y =  -0.716567
        MD2_Normals(92).z =  -0.681718
        MD2_Normals(93).x =    0.147621
        MD2_Normals(93).y =  -0.716567
        MD2_Normals(93).z =  -0.681718
        MD2_Normals(94).x =    0.000000
        MD2_Normals(94).y =  -0.525731
        MD2_Normals(94).z =  -0.850651
        MD2_Normals(95).x =    0.309017
        MD2_Normals(95).y =  -0.500000
        MD2_Normals(95).z =  -0.809017
        MD2_Normals(96).x =    0.442863
        MD2_Normals(96).y =  -0.238856
        MD2_Normals(96).z =  -0.864188
        MD2_Normals(97).x =    0.162460
        MD2_Normals(97).y =  -0.262866
        MD2_Normals(97).z =  -0.951056
        MD2_Normals(98).x =    0.238856
        MD2_Normals(98).y =  -0.864188
        MD2_Normals(98).z =  -0.442863
        MD2_Normals(99).x =    0.500000
        MD2_Normals(99).y =  -0.809017
        MD2_Normals(99).z =  -0.309017
        MD2_Normals(100).x =    0.425325
        MD2_Normals(100).y =  -0.688191
        MD2_Normals(100).z =  -0.587785
        MD2_Normals(101).x =    0.716567
        MD2_Normals(101).y =  -0.681718
        MD2_Normals(101).z =  -0.147621
        MD2_Normals(102).x =    0.688191
        MD2_Normals(102).y =  -0.587785
        MD2_Normals(102).z =  -0.425325
        MD2_Normals(103).x =    0.587785
        MD2_Normals(103).y =  -0.425325
        MD2_Normals(103).z =  -0.688191
        MD2_Normals(104).x =    0.000000
        MD2_Normals(104).y =  -0.955423
        MD2_Normals(104).z =  -0.295242
        MD2_Normals(105).x =    0.000000
        MD2_Normals(105).y =  -1.000000
        MD2_Normals(105).z =   0.000000
        MD2_Normals(106).x =    0.262866
        MD2_Normals(106).y =  -0.951056
        MD2_Normals(106).z =  -0.162460
        MD2_Normals(107).x =    0.000000
        MD2_Normals(107).y =  -0.850651
        MD2_Normals(107).z =   0.525731
        MD2_Normals(108).x =    0.000000
        MD2_Normals(108).y =  -0.955423
        MD2_Normals(108).z =   0.295242
        MD2_Normals(109).x =    0.238856
        MD2_Normals(109).y =  -0.864188
        MD2_Normals(109).z =   0.442863
        MD2_Normals(110).x =    0.262866
        MD2_Normals(110).y =  -0.951056
        MD2_Normals(110).z =   0.162460
        MD2_Normals(111).x =    0.500000
        MD2_Normals(111).y =  -0.809017
        MD2_Normals(111).z =   0.309017
        MD2_Normals(112).x =    0.716567
        MD2_Normals(112).y =  -0.681718
        MD2_Normals(112).z =   0.147621
        MD2_Normals(113).x =    0.525731
        MD2_Normals(113).y =  -0.850651
        MD2_Normals(113).z =   0.000000
        MD2_Normals(114).x =   -0.238856
        MD2_Normals(114).y =  -0.864188
        MD2_Normals(114).z =  -0.442863
        MD2_Normals(115).x =   -0.500000
        MD2_Normals(115).y =  -0.809017
        MD2_Normals(115).z =  -0.309017
        MD2_Normals(116).x =   -0.262866
        MD2_Normals(116).y =  -0.951056
        MD2_Normals(116).z =  -0.162460
        MD2_Normals(117).x =   -0.850651
        MD2_Normals(117).y =  -0.525731
        MD2_Normals(117).z =   0.000000
        MD2_Normals(118).x =   -0.716567
        MD2_Normals(118).y =  -0.681718
        MD2_Normals(118).z =  -0.147621
        MD2_Normals(119).x =   -0.716567
        MD2_Normals(119).y =  -0.681718
        MD2_Normals(119).z =   0.147621
        MD2_Normals(120).x =   -0.525731
        MD2_Normals(120).y =  -0.850651
        MD2_Normals(120).z =   0.000000
        MD2_Normals(121).x =   -0.500000
        MD2_Normals(121).y =  -0.809017
        MD2_Normals(121).z =   0.309017
        MD2_Normals(122).x =   -0.238856
        MD2_Normals(122).y =  -0.864188
        MD2_Normals(122).z =   0.442863
        MD2_Normals(123).x =   -0.262866
        MD2_Normals(123).y =  -0.951056
        MD2_Normals(123).z =   0.162460
        MD2_Normals(124).x =   -0.864188
        MD2_Normals(124).y =  -0.442863
        MD2_Normals(124).z =   0.238856
        MD2_Normals(125).x =   -0.809017
        MD2_Normals(125).y =  -0.309017
        MD2_Normals(125).z =   0.500000
        MD2_Normals(126).x =   -0.688191
        MD2_Normals(126).y =  -0.587785
        MD2_Normals(126).z =   0.425325
        MD2_Normals(127).x =   -0.681718
        MD2_Normals(127).y =  -0.147621
        MD2_Normals(127).z =   0.716567
        MD2_Normals(128).x =   -0.442863
        MD2_Normals(128).y =  -0.238856
        MD2_Normals(128).z =   0.864188
        MD2_Normals(129).x =   -0.587785
        MD2_Normals(129).y =  -0.425325
        MD2_Normals(129).z =   0.688191
        MD2_Normals(130).x =   -0.309017
        MD2_Normals(130).y =  -0.500000
        MD2_Normals(130).z =   0.809017
        MD2_Normals(131).x =   -0.147621
        MD2_Normals(131).y =  -0.716567
        MD2_Normals(131).z =   0.681718
        MD2_Normals(132).x =   -0.425325
        MD2_Normals(132).y =  -0.688191
        MD2_Normals(132).z =   0.587785
        MD2_Normals(133).x =   -0.162460
        MD2_Normals(133).y =  -0.262866
        MD2_Normals(133).z =   0.951056
        MD2_Normals(134).x =    0.442863
        MD2_Normals(134).y =  -0.238856
        MD2_Normals(134).z =   0.864188
        MD2_Normals(135).x =    0.162460
        MD2_Normals(135).y =  -0.262866
        MD2_Normals(135).z =   0.951056
        MD2_Normals(136).x =    0.309017
        MD2_Normals(136).y =  -0.500000
        MD2_Normals(136).z =   0.809017
        MD2_Normals(137).x =    0.147621
        MD2_Normals(137).y =  -0.716567
        MD2_Normals(137).z =   0.681718
        MD2_Normals(138).x =    0.000000
        MD2_Normals(138).y =  -0.525731
        MD2_Normals(138).z =   0.850651
        MD2_Normals(139).x =    0.425325
        MD2_Normals(139).y =  -0.688191
        MD2_Normals(139).z =   0.587785
        MD2_Normals(140).x =    0.587785
        MD2_Normals(140).y =  -0.425325
        MD2_Normals(140).z =   0.688191
        MD2_Normals(141).x =    0.688191
        MD2_Normals(141).y =  -0.587785
        MD2_Normals(141).z =   0.425325
        MD2_Normals(142).x =   -0.955423
        MD2_Normals(142).y =   0.295242
        MD2_Normals(142).z =   0.000000
        MD2_Normals(143).x =   -0.951056
        MD2_Normals(143).y =   0.162460
        MD2_Normals(143).z =   0.262866
        MD2_Normals(144).x =   -1.000000
        MD2_Normals(144).y =   0.000000
        MD2_Normals(144).z =   0.000000
        MD2_Normals(145).x =   -0.850651
        MD2_Normals(145).y =   0.000000
        MD2_Normals(145).z =   0.525731
        MD2_Normals(146).x =   -0.955423
        MD2_Normals(146).y =  -0.295242
        MD2_Normals(146).z =   0.000000
        MD2_Normals(147).x =   -0.951056
        MD2_Normals(147).y =  -0.162460
        MD2_Normals(147).z =   0.262866
        MD2_Normals(148).x =   -0.864188
        MD2_Normals(148).y =   0.442863
        MD2_Normals(148).z =  -0.238856
        MD2_Normals(149).x =   -0.951056
        MD2_Normals(149).y =   0.162460
        MD2_Normals(149).z =  -0.262866
        MD2_Normals(150).x =   -0.809017
        MD2_Normals(150).y =   0.309017
        MD2_Normals(150).z =  -0.500000
        MD2_Normals(151).x =   -0.864188
        MD2_Normals(151).y =  -0.442863
        MD2_Normals(151).z =  -0.238856
        MD2_Normals(152).x =   -0.951056
        MD2_Normals(152).y =  -0.162460
        MD2_Normals(152).z =  -0.262866
        MD2_Normals(153).x =   -0.809017
        MD2_Normals(153).y =  -0.309017
        MD2_Normals(153).z =  -0.500000
        MD2_Normals(154).x =   -0.681718
        MD2_Normals(154).y =   0.147621
        MD2_Normals(154).z =  -0.716567
        MD2_Normals(155).x =   -0.681718
        MD2_Normals(155).y =  -0.147621
        MD2_Normals(155).z =  -0.716567
        MD2_Normals(156).x =   -0.850651
        MD2_Normals(156).y =   0.000000
        MD2_Normals(156).z =  -0.525731
        MD2_Normals(157).x =   -0.688191
        MD2_Normals(157).y =   0.587785
        MD2_Normals(157).z =  -0.425325
        MD2_Normals(158).x =   -0.587785
        MD2_Normals(158).y =   0.425325
        MD2_Normals(158).z =  -0.688191
        MD2_Normals(159).x =   -0.425325
        MD2_Normals(159).y =   0.688191
        MD2_Normals(159).z =  -0.587785
        MD2_Normals(160).x =   -0.425325
        MD2_Normals(160).y =  -0.688191
        MD2_Normals(160).z =  -0.587785
        MD2_Normals(161).x =   -0.587785
        MD2_Normals(161).y =  -0.425325
        MD2_Normals(161).z =  -0.688191
        MD2_Normals(162).x =   -0.688191
        MD2_Normals(162).y =  -0.587785
        MD2_Normals(162).z =  -0.425325

        ' -- Reverting normals
        REGISTER i AS LONG
        FOR i = LBOUND(MD2_Normals) TO UBOUND(MD2_Normals)
          MD2_Normals(i).x *= -1
          MD2_Normals(i).y *= -1
          MD2_Normals(i).z *= -1
        NEXT

        FUNCTION = 1   'success!

        'FUNCTION = 0   'failure!  This will prevent the EXE from running.

    CASE %DLL_PROCESS_DETACH
        'Indicates that the DLL is being unloaded or detached from the
        'calling application.  DLLs can take this opportunity to clean
        'up all resources for all threads attached and known to the DLL.

        FUNCTION = 1   'success!

        'FUNCTION = 0   'failure!

    CASE %DLL_THREAD_ATTACH
        'Indicates that the DLL is being loaded by a new thread in the
        'calling application.  DLLs can use this opportunity to
        'initialize any thread local storage (TLS).

        FUNCTION = 1   'success!

        'FUNCTION = 0   'failure!

    CASE %DLL_THREAD_DETACH
        'Indicates that the thread is exiting cleanly.  If the DLL has
        'allocated any thread local storage, it should be released.

        FUNCTION = 1   'success!

        'FUNCTION = 0   'failure!

    END SELECT

END FUNCTION

' =============================================================================
' -- Auxiliary routines

FUNCTION File_Load( fName AS STRING ) AS STRING
  DIM f AS INTEGER
  DIM sBuffer AS STRING

  f = FREEFILE

  OPEN fName FOR BINARY AS #f
    GET$ #f, LOF(f), sBuffer
  CLOSE #f

  FUNCTION = sBuffer
END FUNCTION

FUNCTION TextureLoadBMP( TexName AS STRING,  TexNum AS LONG, w AS LONG, h AS LONG ) AS LONG

  REGISTER i AS LONG, j AS LONG

  LOCAL n, nWidth, nHeight, nDepth, nTexSize, nColors, diff AS LONG
  LOCAL colors() AS BGRA_t
  LOCAL nTexData, sTex, DescriptionCandidate, WholeBMP AS STRING
  LOCAL pTexData, pTex AS BYTE PTR

  ' -- Determine whether it is physical file or memory image
  IF ISFILE(TexName) = 0 THEN
    MD2_Error( "Texture not found"+TexName )
    FUNCTION = -1
    EXIT FUNCTION
  END IF

  ' Loading file
  n = FREEFILE

  OPEN TexName FOR BINARY AS #n
    GET$ #n, LOF(n), WholeBMP
  CLOSE #n


  DIM bmfh(0 TO 0) AS BitmapFileHeader AT STRPTR(WholeBMP)

  IF bmfh(0).bfType <> 19778 THEN
    TextureLoadingError:
    MD2_Error( "Bad texture format"+TexName )
    FUNCTION = -1
    EXIT FUNCTION
  END IF

  DIM bmih(0 TO 0) AS BitmapInfoHeader AT STRPTR(WholeBMP)+SIZEOF(BitmapFileHeader)

  nWidth  = bmih(0).biWidth
  nHeight = bmih(0).biHeight

  ' > return Back
  w = nWidth
  h = nHeight
  ' <

  nDepth  = bmih(0).biBitCount

  ' Calculate the size of our texture
  nTexSize = nWidth * nHeight * (nDepth \ 8)
  ' Calculate the number of available colors
  nColors = 1
  SHIFT SIGNED LEFT nColors, nDepth
  IF nDepth < 8 THEN
    GOTO TextureLoadingError
  END IF

  ' Load the palette for 8 bits per pixel
  IF nDepth = 8 THEN
    REDIM colors(nColors) AT STRPTR(WholeBMP)+SIZEOF(BitmapFileHeader)+SIZEOF(BitmapInfoHeader)
    nTexData = MID$(WholeBMP, 1+SIZEOF(BitmapFileHeader)+SIZEOF(BitmapInfoHeader)+nColors*4  , LEN(WholeBMP))

  ELSE
    nTexData = MID$(WholeBMP, 1+SIZEOF(BitmapFileHeader)+SIZEOF(BitmapInfoHeader)  , LEN(WholeBMP))
  END IF

  ' --

  ' Calculate the size of this texture once converted to 32-bit RGBA
  diff = nWidth * nHeight * 4

  ' Change format from BGRA to RGBA
  IF nDepth = 8 THEN
    ' Point to the data we loaded in
    pTexData = STRPTR(nTexData)

    ' Allocate memory for our texture
    sTex = SPACE$(diff)
    pTex = STRPTR(sTex)
    IF nHeight > 0 THEN
      j = 0
      ' Count backwards so you start at the front on the texture

      FOR i = 0 TO diff - 1 STEP 4
        ' Transfer the data
        @pTex[i    ] = colors(@pTexData[j]).r
        @pTex[i + 1] = colors(@pTexData[j]).g
        @pTex[i + 2] = colors(@pTexData[j]).b
        @pTex[i + 3] = 255
        INCR j
      NEXT i

    ELSE
      ' Texture parser for a forward texture
      j = nTexSize - 1

      ' Count backwards so you start at the front of the texture
      FOR i = 0 TO diff - 1 STEP 4
          ' Transfer the data
          @pTex[i ] = colors(@pTexData[j]).r
          @pTex[i + 1] = colors(@pTexData[j]).g
          @pTex[i + 2] = colors(@pTexData[j]).b
          @pTex[i + 3] = 255

          DECR j
      NEXT i
    END IF

  ELSE

    ' Allocate memory for our texture
    pTexData = STRPTR(nTexData)
    sTex = SPACE$(diff)
    pTex = STRPTR(sTex): IF pTex = 0 THEN EXIT FUNCTION

    IF nHeight > 0 THEN
      j = 0

      ' Count backwards so you start at the front of the texture
      FOR i = 0 TO nTexSize - 1 STEP 3
        ' Transfer the data
        @pTex[j ] = @pTexData[i + 2]
        @pTex[j + 1] = @pTexData[i + 1]
        @pTex[j + 2] = @pTexData[i ]
        @pTex[j + 3] = 255
        j = j + 4
      NEXT i
    ELSE
      ' Texture parser for a forward texture
      j = diff - 4

      ' Count backwards so you start at the front of the texture
      FOR i = 0 TO nTexSize - 1 STEP 3
          ' Transfer the data
          @pTex[j ] = @pTexData[i + 2]
          @pTex[j + 1] = @pTexData[i + 1]
          @pTex[j + 2] = @pTexData[i ]
          @pTex[j + 3] = 255

          j = j - 4
      NEXT i
    END IF
  END IF

  createGLTexture( pTex , %GL_RGBA8, nWidth, nHeight, TexNum )

  FUNCTION = 0

END FUNCTION

SUB createGLTexture( BYVAL rawBuffer AS BYTE PTR, dataType AS LONG, xSize AS LONG, ySize AS LONG, TexNum AS LONG )

  glBindTexture %GL_TEXTURE_2D, texNum

  glTexParameteri %GL_TEXTURE_2D, %GL_TEXTURE_MAG_FILTER, %GL_LINEAR
  glTexParameteri %GL_TEXTURE_2D, %GL_TEXTURE_MIN_FILTER, %GL_LINEAR_MIPMAP_LINEAR
  gluBuild2DMipmaps %GL_TEXTURE_2D, dataType, xSize, ySize, %GL_RGBA, %GL_UNSIGNED_BYTE, @rawBuffer

END SUB
